<?php
// Copyright (C) 2010 Combodo SARL
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; version 3 of the License.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


/**
 * Module itop-simple-email-synchro: Creation of User request tickets from emails
 *
 * @author      Erwan Taloc <erwan.taloc@combodo.com>
 * @author      Romain Quetiez <romain.quetiez@combodo.com>
 * @author      Denis Flaven <denis.flaven@combodo.com>
 * @license     http://www.opensource.org/licenses/gpl-3.0.html GPL
 */


class iTopSimpleEmailProcessor extends EmailProcessor
{
	/**
	 * Construct a new EmailProcessor... some initialization, reading configuration parameters
	 */
	public function __construct()
	{
		$this->bDebug = MetaModel::GetModuleSetting('itop-simple-email-synchro', 'debug', false);
		$this->aExcludeAttachments = MetaModel::GetModuleSetting('itop-simple-email-synchro', 'exclude_attachment_types', array());
		$this->aDefaultValues = MetaModel::GetModuleSetting('itop-simple-email-synchro', 'default_values', array());		
	}
	
	/**
	 * Outputs some debug text if debugging is enabled from the configuration
	 * @param string $sText The text to output
	 * @return void
	 */
	protected function Trace($sText)
	{
		if ($this->bDebug)
		{
			echo $sText."\n";
		}
	}
	/**
	 * Initializes the email sources... in our case: just one POP3 mailbox
	 * @param void
	 * @return array An array of EmailSource objects
	 */
	public function ListEmailSources()
	{
		$sPop3Server = MetaModel::GetModuleSetting('itop-simple-email-synchro', 'pop3_server', '');
		$sLogin = MetaModel::GetModuleSetting('itop-simple-email-synchro', 'pop3_login', '');
		$sPop3Pwd = MetaModel::GetModuleSetting('itop-simple-email-synchro', 'pop3_pwd', '');
		$iPop3Port = MetaModel::GetModuleSetting('itop-simple-email-synchro', 'pop3_port', 110);
		$pop3AuthOption = MetaModel::GetModuleSetting('itop-simple-email-synchro', 'pop3_auth_option', 'USER');

		$this->Trace("POP3 server: $sPop3Server, POP3 port: $iPop3Port, POP3 login: $sLogin, POP3 password: $sPop3Pwd, POP3 Authentication option: $pop3AuthOption");
		$aSources = array();
		try
		{
			$oSource = new POP3EmailSource($sPop3Server, $iPop3Port, $sLogin, $sPop3Pwd, $pop3AuthOption);
			$aSources[] = $oSource;
		}
		catch(Exception $e)
		{
			$this->Trace("Failed to initialize the mailbox: $sLogin, the mailbox will not be polled. Reason (".$e->getMessage().")");
		}

		return $aSources;
	}
	
	/**
	 * Decides whether a message should be downloaded and processed, deleted, or simply ignored
	 * (i.e left as-is in the mailbox)
	 */
	public function DispatchMessage(EmailSource $oSource, $index, $sUIDL, $oEmailReplica = null)
	{
		$this->Trace("iTop Simple Email Synchro: dispatch of the message $index ($sUIDL)");
		$iRetCode = EmailProcessor::NO_ACTION; // By default, do nothing on this message
		$sRetCode = 'NO_ACTION';
		if ($oEmailReplica == null)
		{
			// New (unread) message
			$this->Trace("The message is new (unread).");
			$iRetCode = EmailProcessor::PROCESS_MESSAGE;
			$sRetCode = 'PROCESS_MESSAGE';
		}
		else
		{
			$iTicketId = $oEmailReplica->Get('ticket_id');
			$oTicket = MetaModel::GetObject('Ticket', $iTicketId, false /* => may not exist anymore */);
			if (is_object($oTicket))
			{
				if ($oTicket->Get('status') == 'closed')
				{
					// The ticket was closed, delete the email (and the replica)
					$this->Trace("The ticket ($iTicketId) corresponding to the message is already closed.");
					$iRetCode = EmailProcessor::DELETE_MESSAGE;
					$sRetCode = 'DELETE_MESSAGE';
				}
				else
				{
					$this->Trace("The ticket ($iTicketId) corresponding to the message exists and is NOT closed.");
				}
			}
			else
			{
				// The ticket was deleted, delete the email (and the replica) as well
				$this->Trace("The ticket ($iTicketId) corresponding to the message no longer exists.");
				$iRetCode = EmailProcessor::DELETE_MESSAGE;
				$sRetCode = 'DELETE_MESSAGE';
			}
		}
		$this->Trace("iTop Simple Email Synchro: dispatch of the message $index ($sUIDL) returned $iRetCode ($sRetCode)");
		return $iRetCode;
	}

	/**
	 * Process the email downloaded from the mailbox.
	 * This implementation creates a UserRequest ticket for each new email.
	 * The caller (identified by its email) must already exists in the database
	 * @param EmailSource $oSource The source from which the email was read
	 * @param integer $index The index of the message in the mailbox
	 * @param EmailMessage $oEmail The downloaded/decoded email message
	 * @param EmailReplica $oEmailReplica The information associating a ticket to the email. Null for new emails
	 */
	public function ProcessMessage(EmailSource $oSource, $index, EmailMessage $oEmail, $oEmailReplica = null)
	{
		$this->Trace("iTop Simple Email Synchro: Processing message $index ({$oEmail->sUIDL})");
		if ($oEmailReplica == null)
		{
			$sContactQuery = 'SELECT Contact WHERE email = :email';
			$oSet = new DBObjectSet(DBObjectSearch::FromOQL($sContactQuery), array(), array('email' => $oEmail->sCallerEmail));
			$sAdditionalDescription = '';
			switch($oSet->Count())
			{
				case 1:
				// Ok, the caller was found in iTop
				$oCaller = $oSet->Fetch();
				break;
				
				case 0:
				// Here decide what to do ? Create a new user, assign the ticket to a 'fixed' Unknown Caller, reject the ticket...
				// For now: let's do nothing, just ignore the ticket, it will be processed again later... in case the caller gets created
				$this->Trace('No contact found for the email address ('.$oEmail->sCallerEmail.'), the ticket will NOT be created');
				return EmailProcessor::NO_ACTION;
				break;
				
				default:
				$this->Trace('Found '.$oSet->Count().' callers with the same email address ('.$oEmail->sCallerEmail.'), the first one will be used...');
				// Multiple callers with the same email address !!!
			 	$sAdditionalDescription = "Warning:\nThere are ".$oSet->Count()." callers with the same email address (".$oEmail->sCallerEmail.") ";
				$sAdditionalDescription = ".\nThe ticket was assigned to the first one found (at random).\n\n";
				$oCaller = $oSet->Fetch();
			}
			
			$oTicket = new UserRequest();
			$oTicket->Set('org_id', $oCaller->Get('org_id'));
			$oTicket->Set('caller_id', $oCaller->GetKey());
			$oTicket->Set('title',$oEmail->sSubject);
			$this->Trace("Email body format: ".$oEmail->sBodyFormat);
			if ($oEmail->sBodyFormat == 'text/html')
			{
				$this->Trace("Removing HTML tags...");
				$sBodyText = $oEmail->StripTags();
			}
			else
			{
				$sBodyText = $oEmail->sBodyText;
			}
			$oTicket->Set('description', $sAdditionalDescription.$sBodyText);
			
			foreach($this->aDefaultValues as $sAttCode => $value)
			{
				$this->InitFromDefaultValue($oTicket, $sAttCode, $value);
			}
			
			$oMyChange = MetaModel::NewObject("CMDBChange");
			$oMyChange->Set("date", time());
			$sUserString = CMDBChange::GetCurrentUserName();
			$oMyChange->Set("userinfo", $sUserString);
			$iChangeId = $oMyChange->DBInsert();
			$oTicket->DBInsertTracked($oMyChange);
			
			// Process attachments (if any)
			foreach($oEmail->aAttachments as $aAttachment)
			{
				// First check if the type is allowed as an attachment...
				if (!in_array($aAttachment['mimeType'], $this->aExcludeAttachments))
				{
					// TODO: allow a scan by an antivirus before uploading to iTop ??
					$oAttachment = new Attachment;
					$oAttachment->Set('item_class', get_class($oTicket));
					$oAttachment->Set('item_id', $oTicket->GetKey());
					$oBlob = new ormDocument($aAttachment['content'], $aAttachment['mimeType'], $aAttachment['filename']);
					$oAttachment->Set('contents', $oBlob);
					$oAttachment->DBInsert();
				}
				else
				{
					$this->Trace("Info: The attachment {$aAttachment['filename']} was NOT added to the ticket because its type '{$aAttachment['mimeType']}' is excluded according to the configuration");
				}
			}

			// Create a replica to keep track that we've processed this email
			$oEmailReplica = new EmailReplica();
			$oEmailReplica->Set('uidl', $oEmail->sUIDL);	
			$oEmailReplica->Set('message_id', $oEmail->sMessageId);
			$oEmailReplica->Set('ticket_id', $oTicket->GetKey());
			$oEmailReplica->DBInsert();
		}
		$this->Trace("iTop Simple Email Synchro: End of processing of the new message $index ({$oEmail->sUIDL}) retCode: NO_ACTION");
		return EmailProcessor::NO_ACTION;	
	}
	
	public function OnDecodeError(EmailSource $oSource, $index, EmailMessage $oEmail)
	{
		
	}
	
	/**
	 * Initializes the ticket from the values specified in the configuration
	 * Each default value must be a valid value for the given field
	 * @param Ticket $oTicket The ticket to update
	 * @param string $sAttCode The code of the attribute to update
	 * @param mixed $value The value to set, for an external key, this is the name of the pointed object
	 */
	protected function InitFromDefaultValue($oTicket, $sAttCode, $value)
	{
		if (!MetaModel::IsValidAttCode(get_class($oTicket), $sAttCode))
	 	{
 			$this->Trace("Warning: cannot set default value '$value'; '$sAttCode' is not a valid attribute of the class ".get_class($oTicket).".");		 		
	 	}
	 	else
	 	{
		 	$oAttDef = MetaModel::GetAttributeDef(get_class($oTicket), $sAttCode);
		 	if (!$oAttDef->IsWritable())
		 	{
		 		$this->Trace("Warning: cannot set default value '$value' for the non-writable attribute: '$sAttCode'.");		 		
		 	}
		 	else
		 	{
		 		$aArgs = array('this' => $oTicket->ToArgs());
		 		$aValues = $oAttDef->GetAllowedValues($aArgs);
		 		if ($aValues == null)
		 		{
		 			// No special constraint for this attribute
			 		if ($oAttDef->IsExternalKey())
			 		{
			 			$oTarget = MetaModel::GetObjectByName($oAttDef->GetTargetClass(), $value, false);
			 			if (is_object($oTarget))
			 			{
			 				$oTicket->Set($sAttCode, $oTarget->GetKey());
			 			}
			 			else
			 			{
				 			$this->Trace("Warning: cannot set default value '$value' for the external key: '$sAttCode'. Unable to find an object of class ".$oAttDef->GetTargetClass()." named '$value'.");
			 			}
			 		}
			 		else if($oAttDef->IsScalar())
			 		{
			 			$oTicket->Set($sAttCode, $value);
			 		}
			 		else
			 		{
			 			$this->Trace("Warning: cannot set default value '$value' for the non-scalar attribute: '$sAttCode'.");
			 		}
		 		}
		 		else
		 		{
		 			// Check that the specified value is a possible/allowed value
			 		if ($oAttDef->IsExternalKey())
			 		{
			 			$bFound = false;
			 			foreach($aValues as $id => $sName)
			 			{
			 				if (strcasecmp($sName,$value) == 0)
			 				{
			 					$bFound = true;
			 					$oTicket->Set($sAttCode, $id);
			 					break;
			 				}
			 			}
			 		}
			 		else if($oAttDef->IsScalar())
			 		{
			 			foreach($aValues as $allowedValue)
			 			{
			 				if ($allowedValue == $value)
			 				{
			 					$bFound = true;
			 					$oTicket->Set($sAttCode, $value);
			 					break;
			 				}
			 			}
			 		}
			 		else
			 		{
			 			$bFound = true;
			 			$this->Trace("Warning: cannot set default value '$value' for the non-scalar attribute: '$sAttCode'.");
			 		}
			 		
			 		if (!$bFound)
			 		{
			 			$this->Trace("Warning: cannot set the value '$value' for the field $sAttCode of the ticket. '$value' is not a valid value for $sAttCode.");		
			 		}
		 		}
			}
		}
	}
}

// register the background action for asynchronous execution in cron.php
EmailBackgroundProcess::RegisterEmailProcessor('iTopSimpleEmailProcessor');
?>